/*======================================================================*/
/*  CopyCons.cpp - see where the Copy Constructor gets called by C++    */
/*======================================================================*/

#include <iostream.h>

class DynInt
{
    public:
        DynInt (void);                      // default constructor
        DynInt (const int nData);           // nData constructor
        DynInt (const DynInt& aDynInt);     // copy constructor
        ~DynInt (void);                     // destructor

        DynInt *DblPtr (void);              // return a pointer
        DynInt& DblRef (void);              // return a reference
        DynInt  DblObj (void);              // return an object

        void PassByValue (DynInt aDynInt);  // pass by value
        void PassByRef (DynInt& aDynInt);   // pass by reference

        unsigned U (void) const;            // pnData as an unsigned int
        friend ostream& operator << (ostream& outs, const DynInt& nDynInt);

    private:
        int  *pnData;                       // the dynamic integer
};

/*======================================================================*/

DynInt::DynInt (void)                       // default constructor
{
    cout << "Default constructor " << U() << endl;
    pnData = NULL;
}

/*======================================================================*/

DynInt::DynInt (const int nData)            // nData constructor
{
    pnData = new int;
    cout << "nData constructor (" << nData << ") " << U() << endl;
    *pnData = nData;
}

/*======================================================================*/

DynInt::DynInt (const DynInt& aDynInt)      // copy constructor
{
    pnData = new int;
    cout << "Copy constructor (" << *aDynInt.pnData << ") " << U() << endl;
    *pnData = *aDynInt.pnData;
}

/*======================================================================*/

DynInt::~DynInt (void)                      // destructor
{
    cout << "Destructor: ";
    if (pnData)
        cout << *pnData << " " << U() << endl;
    else
        cout << "NULL " << U() << endl;

    delete pnData;
    pnData = NULL;
}

/*======================================================================*/

DynInt *DynInt::DblPtr (void)               // return a Pointer
{
    DynInt *pDI = new DynInt (2 * *pnData);
    return pDI;
}

/*======================================================================*/

DynInt& DynInt::DblRef (void)               // return a Reference
{
    DynInt aDynInt (2 * *pnData);
    return aDynInt;
}

/*======================================================================*/

DynInt DynInt::DblObj (void)                // return an Object
{
    DynInt aDynInt (2 * *pnData);
    return aDynInt;
}

/*======================================================================*/

void DynInt::PassByValue (DynInt aDynInt)   // pass by value
{
    cout << "PassByValue = " << aDynInt << endl;
}

/*======================================================================*/

void DynInt::PassByRef (DynInt& aDynInt)    // pass by reference
{
    cout << "PassByRef = " << aDynInt << endl;
}

/*======================================================================*/

unsigned DynInt::U (void) const             // convert Pointer to unsigned
{
    return unsigned (pnData);
}

/*======================================================================*/

ostream& operator << (ostream& outs, const DynInt& aDynInt) // output
{
    if (aDynInt.pnData)
        outs << *aDynInt.pnData;
    else
        outs << "NULL";

    return outs;
}

/*======================================================================*/


int main (void)  {

    cout << endl << "Looking at the Copy Constructor..." << endl << endl;

    cout << "Declaring variables..." << endl;
    DynInt aDI1;
    DynInt aDI2 (123);
    DynInt aDI3 (aDI2);

    cout << endl << "Let's see what they contain..." << endl;
    cout << "aDI1 = " << aDI1 << endl;
    cout << "aDI2 = " << aDI2 << endl;
    cout << "aDI3 = " << aDI3 << endl;

    cout << endl << "Create a pointer to a DI..." << endl;
    DynInt *pDI4 = aDI2.DblPtr ();
    cout << "*pDI4 = " << *pDI4 << endl;
    delete pDI4;
    pDI4 = NULL;

    cout << endl << "Assignment:  call a function returning a Reference" << endl;
    DynInt aDI5 = aDI2.DblRef ();
    cout << "aDI5 = " << aDI5 << endl;

    cout << endl << "Assignment:  call a function returning a Pointer" << endl;
    DynInt *pDI6 = aDI2.DblPtr ();
    cout << "*pDI6 = " << *pDI6 << endl;
    delete pDI6;
    pDI6 = NULL;
    
    cout << endl << "Assignment:  call a function returning an Object" << endl;
    DynInt aDI7 = aDI2.DblObj ();
    cout << "aDI7 = " << aDI7 << endl;

    cout << endl << "Pass by Value" << endl;
    aDI1.PassByValue (aDI3);

    cout << endl << "Pass by Reference" << endl;
    aDI1.PassByRef (aDI3);


    cout << endl << "End of MAIN" << endl << endl;
    return 0;
}

/*======================================================================*/
/*----- output from the above program...


Looking at the Copy Constructor...

Declaring variables...
Default constructor 4204704
nData constructor (123) 7800832
Copy constructor (123) 7800880

Let's see what they contain...
aDI1 = NULL
aDI2 = 123
aDI3 = 123

Create a pointer to a DI...
nData constructor (246) 7800976
*pDI4 = 246
Destructor: 246 7800976

Assignment:  call a function returning a Reference
nData constructor (246) 7801024
Destructor: 246 7801024
Copy constructor (6618680) 7801072
aDI5 = 6618680

Assignment:  call a function returning a Pointer
nData constructor (246) 7801168
*pDI6 = 246
Destructor: 246 7801168

Assignment:  call a function returning an Object
nData constructor (246) 7801216
Copy constructor (246) 7801264
Destructor: 246 7801216
aDI7 = 246

Pass by Value
Copy constructor (123) 7801312
PassByValue = 123
Destructor: 123 7801312

Pass by Reference
PassByRef = 123

End of MAIN

Destructor: 246 7801264
Destructor: 6618680 7801072
Destructor: 123 7800880
Destructor: 123 7800832
Destructor: NULL 0

-----*/

